//Author: Max Krasnow
//Collaborators: Andy Delton, John Tooby, Leda Cosmides
//Compiled: August 8, 2011

/*
Copyright  2011.  The Regents of the University of California (Regents).
All rights reserved.
Redistribution and use in source and binary forms, with or without modification, are permitted for non-commercial educational, non-commercial research, and not-for-profit purposes only, provided that the following conditions are met:
	Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
	Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
	Neither the name of The Regents or the University of California nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
package sim1;

import java.util.ArrayList;
import java.util.Collections;

public class population {
    private ArrayList<agent> AgentList;
    private ArrayList<group> GroupList;
    private ArrayList<agent> nextGenList = new ArrayList();
    private int MaxPopSize;
    
    //Creates a new instance of population
    
    public population(int popsize, double startingBC, double mutationsize){
        MaxPopSize = popsize;
        AgentList = new ArrayList();
        for(int i=0; i < Math.floor(popsize); i++){
            BFagent a = new BFagent(startingBC, 1, mutationsize);
            AgentList.add(a);

        }
    }
    public population(int popsize, double startingprCifR, double startingprCifOS, double mutationsize){
        MaxPopSize = popsize;
        AgentList = new ArrayList();
        for(int i=0; i < Math.floor(popsize); i++){
            BIagent a = new BIagent(startingprCifR, startingprCifOS, 1, mutationsize);
            AgentList.add(a);
        }
    }
       
    public population(ArrayList<agent> a){
        AgentList = new ArrayList();
        AgentList=a;
        MaxPopSize = a.size();
    }
    
    public void shuffle(){
        int i = 5 + (int)(Math.random()*5);
        for(int x=0; x<i; x++){
            Collections.shuffle(AgentList);
        }
    }
      
    public int numGroups(){
        return GroupList.size();
    }
    public int numAgents(){
        return AgentList.size();
    }
    
    public agent getAgent(int i){
        return AgentList.get(i);
    }
    
    public group getGroup(int i){
        return GroupList.get(i);
    }
    public void setAgentList(ArrayList<agent> i){
        AgentList=i;
    }
  
    public void formGroups(int n) {
        group temp = new group();
        GroupList = new ArrayList();
        for(int i = 0; i < MaxPopSize; i++){
            if(i>0&&i%n==0){
                GroupList.add(temp);
                temp = new group();
            }
            temp.members.add(AgentList.get(i));
        }
        GroupList.add(temp);
    }
          
    public result beliefFormation(double b, double c, double w, double basefit, double d, double osp, double mutrate, double mutsize, double errorRate){
        //update stamps based on group type (and asssign group types)
        for(int i=0; i < GroupList.size(); i++){
            if(i<GroupList.size()*osp){
                group thisGroup = GroupList.get(i);
                thisGroup.type="OS";
                BFagent a1 = (BFagent) GroupList.get(i).members.get(0);
                BFagent a2 = (BFagent) GroupList.get(i).members.get(1);
                a1.setStamp(a1.getStamp()+d/2);
                a2.setStamp(a2.getStamp()+d/2);
                thisGroup.members.add(a1);
                thisGroup.members.add(a2);
                GroupList.set(i, thisGroup);
            } else {
                group thisGroup = GroupList.get(i);
                thisGroup.type="R";
                BFagent a1 = (BFagent) GroupList.get(i).members.get(0);
                BFagent a2 = (BFagent) GroupList.get(i).members.get(1);
                a1.setStamp(a1.getStamp()-d/2);
                a2.setStamp(a2.getStamp()-d/2);
                thisGroup.members.add(a1);
                thisGroup.members.add(a2);
                GroupList.set(i, thisGroup);
            }
        }
        //play game and assign fitness
        AgentList.clear();
        for(int i=0; i < GroupList.size(); i++){
            group thisGroup = GroupList.get(i);
            BFagent a1 = (BFagent) thisGroup.members.get(0);
            BFagent a2 = (BFagent) thisGroup.members.get(1);
            a1.setFitness(basefit);
            a2.setFitness(basefit);
            double rand=0;
            while(w>rand){
                if(a1.BFcog(a2.getStamp(), errorRate)=='c'){
                    if(a2.BFcog(a1.getStamp(), errorRate)=='c'){
                        a1.BFexperience('c');
                        a1.changeFitness(b-c);
                        a2.BFexperience('c');
                        a2.changeFitness(b-c);
                    } else {
                        a1.BFexperience('d');
                        a1.changeFitness(-c);
                        a2.BFexperience('c');
                        a2.changeFitness(b);
                    }
                } else {
                    if(a2.BFcog(a1.getStamp(), errorRate)=='c'){
                        a1.BFexperience('c');
                        a1.changeFitness(b);
                        a1.BFexperience('d');
                        a1.changeFitness(-c);
                    } else {
                        a1.BFexperience('d');
                        a1.BFexperience('d');
                    }
                }
                if(thisGroup.type.matches("R")){
                    rand=Math.random();
                }
                if(thisGroup.type.matches("OS")){
                    rand=1;
                }
            }
            AgentList.add(a1);
            AgentList.add(a2);
            thisGroup.members.clear();
        }
        reproduce("Cognition Evolves", mutrate, mutsize);
        ArrayList<TypeDistribution> nextGenDist = new ArrayList();
        TypeDistribution BCmean = new TypeDistribution("BC", mean(0), SDSE(0));
        nextGenDist.add(BCmean);
        result thisResult = new result(nextGenList, nextGenDist);
        thisResult.b = b;
        thisResult.c = c;
        thisResult.pop = nextGenList.size();
        thisResult.basefit = basefit;
        thisResult.w = w;
        thisResult.d = d;
        thisResult.osp = osp;
        thisResult.ms = mutsize;
        thisResult.mr = mutrate;
        return thisResult;
    }
    
     public result beliefImplementation(double b, double c, double w, double basefit, double d, double osp, double mutrate, double mutsize, double errorRate){
        //update stamps based on group type (and asssign group types)
        for(int i=0; i < GroupList.size(); i++){
            if(i<GroupList.size()*osp){
                group thisGroup = GroupList.get(i);
                thisGroup.type="OS";
                BIagent a1 = (BIagent) GroupList.get(i).members.get(0);
                BIagent a2 = (BIagent) GroupList.get(i).members.get(1);
                a1.setStamp(a1.getStamp()+d/2);
                a2.setStamp(a2.getStamp()+d/2);
                thisGroup.members.add(a1);
                thisGroup.members.add(a2);
                GroupList.set(i, thisGroup);
            } else {
                group thisGroup = GroupList.get(i);
                thisGroup.type="R";
                BIagent a1 = (BIagent) GroupList.get(i).members.get(0);
                BIagent a2 = (BIagent) GroupList.get(i).members.get(1);
                a1.setStamp(a1.getStamp()-d/2);
                a2.setStamp(a2.getStamp()-d/2);
                thisGroup.members.add(a1);
                thisGroup.members.add(a2);
                GroupList.set(i, thisGroup);
            }
        }
        AgentList.clear();
        for(int i=0; i < GroupList.size(); i++){
            group thisGroup = GroupList.get(i);
            BIagent a1 = (BIagent) thisGroup.members.get(0);
            BIagent a2 = (BIagent) thisGroup.members.get(1);
            a1.setFitness(basefit);
            a2.setFitness(basefit);
            double rand=0;
            while(w>rand){
                if(a1.BIcog(a2.getStamp(), osp, d, errorRate)=='c'){
                    if(a2.BIcog(a1.getStamp(), osp, d, errorRate)=='c'){
                        a1.BIexperience('c');
                        a1.changeFitness(b-c);
                        a2.BIexperience('c');
                        a2.changeFitness(b-c);
                    } else {
                        a1.BIexperience('d');
                        a1.changeFitness(-c);
                        a2.BIexperience('c');
                        a2.changeFitness(b);
                    }
                } else {
                    if(a2.BIcog(a1.getStamp(), osp, d, errorRate)=='c'){
                        a1.BIexperience('c');
                        a1.changeFitness(b);
                        a1.BIexperience('d');
                        a1.changeFitness(-c);
                    } else {
                        a1.BIexperience('d');
                        a1.BIexperience('d');
                    }
                }
                if(thisGroup.type.matches("R")){
                    rand=Math.random();
                }
                if(thisGroup.type.matches("OS")){
                    rand=1;
                }
            }
            thisGroup.members.clear();
            AgentList.add(a1);
            AgentList.add(a2);
        }
        
        //do reproduction functions
        reproduce("Motivation Evolves", mutrate, mutsize);
        ArrayList<TypeDistribution> nextGenDist = new ArrayList();
        TypeDistribution prCifRmean = new TypeDistribution("prCifR", mean(0), SDSE(0));
        TypeDistribution prCifOSmean = new TypeDistribution("prCifOS", mean(1), SDSE(1));
        nextGenDist.add(prCifRmean);
        nextGenDist.add(prCifOSmean);
        result thisResult = new result(nextGenList, nextGenDist);
        thisResult.b = b;
        thisResult.c = c;
        thisResult.pop = nextGenList.size();
        thisResult.basefit = basefit;
        thisResult.w = w;
        thisResult.d = d;
        thisResult.osp = osp;
        thisResult.ms = mutsize;
        thisResult.mr = mutrate;
        return thisResult;
     }
            

     private double count(int i, int j){
         double runTot = 0;
         for(agent thisAgent:nextGenList){
             if(thisAgent.genes.get(i).value==j){
                 runTot+=1;
             }
         }
         runTot/=nextGenList.size();
         return runTot;
     }
     
     private double mean(int i){
         double runTot = 0;
         for(agent thisAgent:nextGenList){
             runTot += thisAgent.genes.get(i).value;
         }
         runTot/=nextGenList.size();

         return runTot;
     }
     
     private double[] SDSE(int i){
         double runTot = 0;
         double runSq = 0;
         for(agent thisAgent:nextGenList){
             runTot += thisAgent.genes.get(i).value;
             runSq += Math.pow(thisAgent.genes.get(i).value, 2);
         }
         double SD = Math.pow((runSq - (Math.pow(runTot,2)/MaxPopSize))/(MaxPopSize-1),0.5);
         double SE = SD/Math.pow(nextGenList.size(),.5);
         double[] toReturn = {SD,SE};
         return toReturn;
     }    
     
     private void reproduce(String game, double mutrate, double mutsize){
        double totalFitness=0;
        double runningProp=0;
        ArrayList<Double> propList = new ArrayList();
        double lowestFitness=0;
        for(agent thisAgent:AgentList){
            if(thisAgent.getFitness()<lowestFitness){
                lowestFitness = thisAgent.getFitness();
            }
        }
        if(lowestFitness<0){
            lowestFitness=Math.abs(lowestFitness)+1;
            for(int x=0; x<MaxPopSize; x++){
                AgentList.get(x).changeFitness(lowestFitness);
            }
        }
        shuffle();
        for(agent thisAgent:AgentList){
            totalFitness+=thisAgent.getFitness();
        }
        for(int i=0; i<MaxPopSize; i++){
            AgentList.get(i).scaleFitness(totalFitness);
            runningProp+=AgentList.get(i).getFitness();
            propList.add(runningProp);
        }
        propList.set(propList.size()-1,1.0);
        for(int i=0; i<MaxPopSize; i++){
            double rnd = Math.random();
            boolean condition = true;
            int z=0;
            while(condition){
                if(rnd<=propList.get(z)){
                    condition=false;
                    if(game.matches("Cognition Evolves")){
                        nextGenList.add(new BFagent(AgentList.get(z).genes.get(0).value, mutrate, mutsize));
                    } else if(game.matches("Motivation Evolves")){
                        nextGenList.add(new BIagent(AgentList.get(z).genes.get(0).value, AgentList.get(z).genes.get(1).value, mutrate, mutsize));
                    } else {
                        System.out.println("No Such Game");
                    }
                } else {
                    z++;
                    if(z>=MaxPopSize){
                        int lastz = z-1;
                        for(double prop:propList){
                            System.out.print(prop + ", ");
                        }
                        System.out.println("Trouble!");
                        break;
                    }
                }
            } 
         }
        AgentList.clear();
     }   
}